"""All decorated callback functions should be defined here.

Implementations of the callback functionality should be in other functions (in other files), to enable unit testing.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from dash import ALL, Dash, Input, Output, State, ctx
from datadoc_model.Enums import SupportedLanguages

from datadoc import state
from datadoc.frontend.callbacks.dataset import (
    accept_dataset_metadata_input,
    change_language_dataset_metadata,
    open_dataset_handling,
)
from datadoc.frontend.callbacks.variables import (
    accept_variable_metadata_input,
    update_variable_table_dropdown_options_for_language,
    update_variable_table_language,
)
from datadoc.frontend.components.dataset_tab import DATASET_METADATA_INPUT
from datadoc.frontend.fields.display_dataset import DISPLAYED_DROPDOWN_DATASET_METADATA

if TYPE_CHECKING:
    from datadoc.frontend.callbacks.utils import MetadataInputTypes


def register_callbacks(app: Dash) -> None:
    """Define and register callbacks."""

    @app.callback(
        Output("progress-bar", "value"),
        Output("progress-bar", "label"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        Input("variables-table", "data"),
    )
    def callback_update_progress(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
        data: list[dict],  # noqa: ARG001 argument required by Dash
    ) -> tuple[int, str]:
        """Update the progress bar when new data is entered."""
        completion = state.metadata.percent_complete
        return completion, f"{completion}%"

    @app.callback(
        Output("saved-metadata-success", "is_open"),
        Input("save-button", "n_clicks"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(n_clicks: int) -> bool:
        """Save the metadata document to disk."""
        if n_clicks and n_clicks > 0:
            # Write the final completion percentage to the model
            state.metadata.meta.percentage_complete = state.metadata.percent_complete
            state.metadata.write_metadata_document()
            return True

        return False

    @app.callback(
        *[
            Output(
                {
                    "type": DATASET_METADATA_INPUT,
                    "id": m.identifier,
                },
                "options",
            )
            for m in DISPLAYED_DROPDOWN_DATASET_METADATA
        ],
        Output(
            {"type": DATASET_METADATA_INPUT, "id": ALL},
            "value",
        ),
        Input("language-dropdown", "value"),
    )
    def callback_change_language_dataset_metadata(
        language: str,
    ) -> tuple[tuple[list[dict[str, str]], ...], list]:
        """Update dataset metadata values upon change of language."""
        return change_language_dataset_metadata(SupportedLanguages(language))

    @app.callback(
        Output("dataset-validation-error", "is_open"),
        Output("dataset-validation-explanation", "children"),
        Input({"type": DATASET_METADATA_INPUT, "id": ALL}, "value"),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
        )

    @app.callback(
        Output("variables-table", "data"),
        Output("variables-validation-error", "is_open"),
        Output("variables-validation-explanation", "children"),
        State("variables-table", "active_cell"),
        Input("variables-table", "data"),
        State("variables-table", "data_previous"),
        Input("language-dropdown", "value"),
        prevent_initial_call=True,
    )
    def callback_variable_table(
        active_cell: dict,
        data: list[dict],
        data_previous: list[dict],
        language: str,
    ) -> tuple[list[dict], bool, str]:
        """Update data in the variable table.

        Triggered upon:
        - New data enetered in the variable table.
        - Change of language.

        Will display an alert if validation fails.
        """
        if ctx.triggered_id == "language-dropdown":
            return update_variable_table_language(SupportedLanguages(language))

        return accept_variable_metadata_input(data, active_cell, data_previous)

    @app.callback(
        Output("variables-table", "dropdown"),
        Input("language-dropdown", "value"),
    )
    def callback_variable_table_dropdown_options(
        language: str,
    ) -> dict[str, dict[str, list[dict[str, str]]]]:
        """Update the options in variable table dropdowns when the language changes."""
        language = SupportedLanguages(language)
        return update_variable_table_dropdown_options_for_language(language)

    @app.callback(
        Output("opened-dataset-success", "is_open"),
        Output("opened-dataset-error", "is_open"),
        Output("opened-dataset-error-explanation", "children"),
        Output("language-dropdown", "value"),  # Used to force reload of metadata
        Input("open-button", "n_clicks"),
        State("dataset-path-input", "value"),
    )
    def callback_open_dataset(
        n_clicks: int,
        dataset_path: str,
    ) -> tuple[bool, bool, str, SupportedLanguages]:
        """Open a dataset.

        Shows an alert on success or failure.

        To trigger reload of data in the UI, we update the
        language dropdown. This is a hack and could be replaced
        by a more formal mechanism.
        """
        return open_dataset_handling(n_clicks, dataset_path)



"""Handle reading, updating and writing of metadata."""
from __future__ import annotations

import json
import logging
import os
import pathlib
import typing as t
import uuid
from typing import TYPE_CHECKING

from datadoc_model import Model
from datadoc_model.Enums import DatasetState, SupportedLanguages

from datadoc.backend.dataset_parser import DatasetParser
from datadoc.backend.model_backwards_compatibility import upgrade_metadata
from datadoc.backend.storage_adapter import StorageAdapter
from datadoc.frontend.fields import display_dataset, display_variables
from datadoc.utils import calculate_percentage, get_timestamp_now

if TYPE_CHECKING:
    from datetime import datetime

logger = logging.getLogger(__name__)

OBLIGATORY_DATASET_METADATA = [
    m.identifier
    for m in display_dataset.DISPLAY_DATASET.values()
    if m.obligatory and m.editable
]

OBLIGATORY_VARIABLES_METADATA = [
    m.identifier
    for m in display_variables.DISPLAY_VARIABLES.values()
    if m.obligatory and m.editable
]

# These don't vary at runtime so we calculate them as constants here
NUM_OBLIGATORY_DATASET_FIELDS = len(
    [k for k in Model.DataDocDataSet().dict() if k in OBLIGATORY_DATASET_METADATA],
)
NUM_OBLIGATORY_VARIABLES_FIELDS = len(
    [k for k in Model.DataDocVariable().dict() if k in OBLIGATORY_VARIABLES_METADATA],
)

METADATA_DOCUMENT_FILE_SUFFIX = "__DOC.json"

PLACEHOLDER_USERNAME = "default_user@ssb.no"


class DataDocMetadata:
    """Handle reading, updating and writing of metadata."""

    def __init__(
        self: t.Self @ DataDocMetadata,
        dataset: str | None,
    ) -> None:
        """Read in a dataset if supplied, otherwise naively instantiate the class."""
        self.dataset: str = dataset
        if self.dataset:
            self.short_name: str = pathlib.Path(
                self.dataset,
            ).stem  # filename without file ending
            self.metadata_document: StorageAdapter = StorageAdapter.for_path(
                StorageAdapter.for_path(self.dataset).parent(),
            )
            self.metadata_document.joinpath(
                self.short_name + METADATA_DOCUMENT_FILE_SUFFIX,
            )
            self.dataset_state: DatasetState = self.get_dataset_state(self.dataset)
        try:
            self.current_user = os.environ["JUPYTERHUB_USER"]
        except KeyError:
            self.current_user = PLACEHOLDER_USERNAME
            logger.warning(
                "JUPYTERHUB_USER env variable not set, using %s as placeholder",
                self.current_user,
            )

        self.meta: Model.MetadataDocument = Model.MetadataDocument(
            percentage_complete=0,
            document_version=Model.MODEL_VERSION,
            dataset=Model.DataDocDataSet(),
            variables=[],
        )

        self.variables_lookup: dict[str, Model.DataDocVariable] = {}

        if self.dataset:
            self.extract_metadata_from_files()

    def get_dataset_state(
        self: t.Self @ DataDocMetadata,
        dataset: str,
    ) -> DatasetState | None:
        """Use the path to attempt to guess the state of the dataset."""
        if dataset is None:
            return None
        dataset_path_parts = set(pathlib.Path(dataset).parts)
        for state in DatasetState:
            # We assume that files are saved in the Norwegian language as specified by SSB.
            norwegian_dataset_state_path_part = state.get_value_for_language(
                SupportedLanguages.NORSK_BOKMÅL,
            ).lower()
            norwegian_dataset_state_path_part_variations = {
                norwegian_dataset_state_path_part.replace(" ", x) for x in ["-", "_"]
            }
            # Match on any of the variations anywhere in the path.
            if norwegian_dataset_state_path_part_variations.intersection(
                dataset_path_parts,
            ):
                return state

        return None

    def get_dataset_version(
        self: t.Self @ DataDocMetadata,
        dataset_stem: str,
    ) -> str | None:
        """Find version information if exists in filename.

        eg. 'v1' in filename 'person_data_v1.parquet'
        """
        minimum_elements_in_file_name: t.Final[int] = 2
        minimum_characters_in_version_string: t.Final[int] = 2
        split_file_name = str(dataset_stem).split("_")
        if len(split_file_name) >= minimum_elements_in_file_name:
            last_filename_element = str(split_file_name[-1])
            if (
                len(last_filename_element) >= minimum_characters_in_version_string
                and last_filename_element[0:1] == "v"
                and last_filename_element[1:].isdigit()
            ):
                return last_filename_element[1:]
        return None

    def extract_metadata_from_files(self: t.Self @ DataDocMetadata) -> None:
        """Read metadata from a dataset.

        If a metadata document already exists, read in the metadata from that instead.
        """
        fresh_metadata = {}
        if self.metadata_document.exists():
            try:
                with self.metadata_document.open(mode="r", encoding="utf-8") as file:
                    fresh_metadata = json.load(file)
                logger.info(
                    "Opened existing metadata file %s",
                    self.metadata_document.location,
                )

                fresh_metadata = upgrade_metadata(fresh_metadata, Model.MODEL_VERSION)

                variables_list = fresh_metadata.pop("variables", None)

                self.meta.variables = [
                    Model.DataDocVariable(**v) for v in variables_list
                ]
                self.meta.dataset = Model.DataDocDataSet(
                    **fresh_metadata.pop("dataset", None),
                )
            except json.JSONDecodeError:
                logger.warning(
                    "Could not open existing metadata file %s. \
                    Falling back to collecting data from the dataset",
                    self.metadata_document.location,
                    exc_info=True,
                )
                self.extract_metadata_from_dataset()
        else:
            self.extract_metadata_from_dataset()

        if self.meta.dataset.id is None:
            self.meta.dataset.id = uuid.uuid4()

        # Set default values for variables where appropriate
        v: Model.DataDocVariable
        for v in self.meta.variables:
            if v.variable_role is None:
                v.variable_role = Model.Enums.VariableRole.MEASURE
            if v.direct_person_identifying is None:
                v.direct_person_identifying = False

        self.variables_lookup = {v.short_name: v for v in self.meta.variables}

    def extract_metadata_from_dataset(self: t.Self @ DataDocMetadata) -> None:
        """Obtain what metadata we can from the dataset itself.

        This makes it easier for the user by 'pre-filling' certain fields.
        Certain elements are dependent on the dataset being saved according to SSB's standard.
        """
        self.ds_schema = DatasetParser.for_file(self.dataset)

        self.meta.dataset = Model.DataDocDataSet(
            short_name=self.short_name,
            dataset_state=self.dataset_state,
            version=self.get_dataset_version(self.short_name),
            data_source_path=self.dataset,
            created_by=self.current_user,
        )

        self.meta.variables = self.ds_schema.get_fields()

    def write_metadata_document(self: t.Self @ DataDocMetadata) -> None:
        """Write all currently known metadata to file."""
        timestamp: datetime = get_timestamp_now()
        if self.meta.dataset.metadata_created_date is None:
            self.meta.dataset.metadata_created_date = timestamp
        if self.meta.dataset.metadata_created_by is None:
            self.meta.dataset.metadata_created_by = self.current_user
        self.meta.dataset.metadata_last_updated_date = timestamp
        self.meta.dataset.metadata_last_updated_by = self.current_user
        self.metadata_document.write_text(self.meta.json(indent=4, sort_keys=False))
        logger.info("Saved metadata document %s", self.metadata_document.location)

    @property
    def percent_complete(self: t.Self @ DataDocMetadata) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator.
        """
        num_all_fields = NUM_OBLIGATORY_DATASET_FIELDS
        num_set_fields = len(
            [
                k
                for k, v in self.meta.dataset.dict().items()
                if k in OBLIGATORY_DATASET_METADATA and v is not None
            ],
        )

        for variable in self.meta.variables:
            num_all_fields += NUM_OBLIGATORY_VARIABLES_FIELDS
            num_set_fields += len(
                [
                    k
                    for k, v in variable.dict().items()
                    if k in OBLIGATORY_VARIABLES_METADATA and v is not None
                ],
            )

        return calculate_percentage(num_set_fields, num_all_fields)



"""Tests for the utils module."""

from datadoc_model.Enums import SupportedLanguages
from datadoc_model.Model import DataDocVariable

from datadoc.tests.test_callbacks import BOKMÅL_NAME, LANGUAGE_OBJECT
from datadoc.utils import calculate_percentage, get_display_values, running_in_notebook


def test_not_running_in_notebook():
    assert not running_in_notebook()


def test_calculate_percentage():
    assert calculate_percentage(1, 3) == 33  # noqa: PLR2004


def test_get_display_values():
    variable = DataDocVariable(name=LANGUAGE_OBJECT)
    values = get_display_values(variable, SupportedLanguages.NORSK_BOKMÅL)
    assert values["name"] == BOKMÅL_NAME



"""Smoke tests."""
import pytest
from dash import Dash
from jupyter_dash import JupyterDash

from datadoc import state
from datadoc.app import build_app
from datadoc.backend.datadoc_metadata import DataDocMetadata

from .utils import TEST_PARQUET_FILEPATH


@pytest.mark.parametrize("dash_class", [Dash, JupyterDash])
def test_build_app_dash(dash_class: Dash):
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    app = build_app(dash_class)
    assert app.config["name"] == "Datadoc"
    assert len(app.callback_map.items()) > 0



"""Upgrade old metadata files to be compatible with new versions of the datadoc model.

An important principle of Datadoc is that we ALWAYS guarantee backwards compatibility of existing metadata documents.
This means that we guarantee that a user will never lose data, even if their document is decades old.

For each document version we release with breaking changes, we implement a handler and register the version by defining a
BackwardsCompatibleVersion instance. These documents will then be upgraded when they're opened in Datadoc.

A test must also be implemented for each new version.
"""

from __future__ import annotations

import typing as t
from dataclasses import dataclass

if t.TYPE_CHECKING:
    from collections.abc import Callable

VERSION_FIELD_NAME = "document_version"


class UnknownModelVersionError(Exception):
    """Throw this error if we haven't seen the version before."""

    def __init__(
        self: t.Self @ UnknownModelVersionError,
        supplied_version: str,
        *args: tuple[t.Any, ...],
    ) -> None:
        """Initialize class."""
        super().__init__(args)
        self.supplied_version = supplied_version

    def __str__(self: t.Self @ UnknownModelVersionError) -> str:
        """Return string representation."""
        return f"Document Version ({self.supplied_version}) of discovered file is not supported"


SUPPORTED_VERSIONS: dict[str, BackwardsCompatibleVersion] = {}


@dataclass()
class BackwardsCompatibleVersion:
    """A version which we support with backwards compatibility."""

    version: str
    handler: Callable

    def __post_init__(self: t.Self @ BackwardsCompatibleVersion) -> None:
        """Register this version in the supported versions map."""
        SUPPORTED_VERSIONS[self.version] = self


def handle_current_version(supplied_metadata: dict) -> dict:
    """Nothing to do here."""
    return supplied_metadata


def handle_version_0_1_1(supplied_metadata: dict) -> dict:
    """Handle breaking changes for v0.1.1.

    PR ref: https://github.com/statisticsnorway/ssb-datadoc-model/pull/4.
    """
    key_renaming = [
        ("metadata_created_date", "created_date"),
        ("metadata_created_by", "created_by"),
        ("metadata_last_updated_date", "last_updated_date"),
        ("metadata_last_updated_by", "last_updated_by"),
    ]
    for new_key, old_key in key_renaming:
        supplied_metadata["dataset"][new_key] = supplied_metadata["dataset"].pop(
            old_key,
        )
    return supplied_metadata


# Register all the supported versions and their handlers
BackwardsCompatibleVersion(version="0.1.1", handler=handle_version_0_1_1)
BackwardsCompatibleVersion(
    version="1",
    handler=handle_version_0_1_1,
)  # Some documents exist with incorrect version specification


def upgrade_metadata(fresh_metadata: dict, current_model_version: str) -> dict:
    """Run the handler for this version to upgrade the document to the latest version."""
    # Special case for current version, we expose the current_model_version parameter for test purposes
    SUPPORTED_VERSIONS[current_model_version] = BackwardsCompatibleVersion(
        current_model_version,
        handle_current_version,
    )
    supplied_version = fresh_metadata[VERSION_FIELD_NAME]
    try:
        # Retrieve the upgrade function for this version
        upgrade = SUPPORTED_VERSIONS[supplied_version].handler
    except KeyError as e:
        raise UnknownModelVersionError(supplied_version) from e
    else:
        return upgrade(fresh_metadata)



"""Verify that we are in sync with the Model."""

from datadoc_model.Model import DataDocDataSet, DataDocVariable

from datadoc.frontend.fields.display_dataset import DISPLAY_DATASET, DatasetIdentifiers
from datadoc.frontend.fields.display_variables import (
    DISPLAY_VARIABLES,
    VariableIdentifiers,
)


def test_dataset_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in DatasetIdentifiers] == list(DataDocDataSet().dict().keys())
    assert list(DatasetIdentifiers) == list(DISPLAY_DATASET.keys())


def test_variables_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    assert [i.value for i in VariableIdentifiers] == list(
        DataDocVariable().dict().keys(),
    )
    assert list(VariableIdentifiers) == list(DISPLAY_VARIABLES.keys())



"""Functions which aren't directly called from a decorated callback."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from datadoc_model import Model

from datadoc import state

if TYPE_CHECKING:
    from datadoc_model.Enums import SupportedLanguages
    from datadoc_model.LanguageStringsEnum import LanguageStringsEnum

logger = logging.getLogger(__name__)


MetadataInputTypes: type = str | int | float | bool | None


def update_global_language_state(language: SupportedLanguages) -> None:
    """Update global language state."""
    logger.debug("Updating language: %s", language.name)
    state.current_metadata_language = language


def get_options_for_language(
    language: SupportedLanguages,
    enum: LanguageStringsEnum,
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    return [
        {
            "label": i.get_value_for_language(language),
            "value": i.name,
        }
        for i in enum
    ]


def find_existing_language_string(
    metadata_model_object: Model.DataDocBaseModel,
    value: str,
    metadata_identifier: str,
) -> Model.LanguageStrings | None:
    """Get or create a LanguageStrings object and return it."""
    # In this case we need to set the string to the correct language code
    language_strings = getattr(metadata_model_object, metadata_identifier)
    if language_strings is None:
        # This means that no strings have been saved yet so we need to construct
        # a new LanguageStrings object
        language_strings = Model.LanguageStrings(
            **{state.current_metadata_language.value: value},
        )
    else:
        # In this case there's an existing object so we save this string
        # to the current language
        setattr(language_strings, state.current_metadata_language.value, value)
    return language_strings



"""Tests for the DatasetParser class."""

import pytest
from datadoc_model.Enums import Datatype, SupportedLanguages
from datadoc_model.Model import DataDocVariable, LanguageStrings

from datadoc import state
from datadoc.backend.dataset_parser import (
    KNOWN_BOOLEAN_TYPES,
    KNOWN_DATETIME_TYPES,
    KNOWN_FLOAT_TYPES,
    KNOWN_INTEGER_TYPES,
    KNOWN_STRING_TYPES,
    DatasetParser,
    DatasetParserParquet,
)

from .utils import (
    TEST_PARQUET_FILEPATH,
    TEST_PARQUET_GZIP_FILEPATH,
    TEST_SAS7BDAT_FILEPATH,
)


def test_use_abstract_class_directly():
    with pytest.raises(TypeError):
        DatasetParser().get_fields()


@pytest.mark.parametrize(
    "local_parser",
    [
        DatasetParser.for_file(TEST_PARQUET_FILEPATH),
        DatasetParser.for_file(TEST_PARQUET_GZIP_FILEPATH),
    ],
)
def test_get_fields_parquet(local_parser: DatasetParserParquet):
    expected_fields = [
        DataDocVariable(short_name="pers_id", data_type=Datatype.STRING),
        DataDocVariable(short_name="tidspunkt", data_type=Datatype.DATETIME),
        DataDocVariable(short_name="sivilstand", data_type=Datatype.STRING),
        DataDocVariable(short_name="alm_inntekt", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="sykepenger", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="ber_bruttoformue", data_type=Datatype.INTEGER),
        DataDocVariable(short_name="fullf_utdanning", data_type=Datatype.STRING),
        DataDocVariable(short_name="hoveddiagnose", data_type=Datatype.STRING),
    ]
    fields = local_parser.get_fields()

    assert fields == expected_fields


def test_get_fields_sas7bdat():
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    expected_fields = [
        DataDocVariable(
            short_name="tekst",
            name=LanguageStrings(nb="Tekst"),
            data_type=Datatype.STRING,
        ),
        DataDocVariable(
            short_name="tall",
            name=LanguageStrings(nb="Tall"),
            data_type=Datatype.FLOAT,
        ),
        DataDocVariable(
            short_name="dato",
            name=LanguageStrings(nb="Dato"),
            data_type=Datatype.DATETIME,
        ),
    ]

    reader = DatasetParser.for_file(TEST_SAS7BDAT_FILEPATH)
    fields = reader.get_fields()

    assert fields == expected_fields


def test_get_fields_unknown_file_type():
    with pytest.raises(NotImplementedError):
        DatasetParser.for_file("my_dataset.csv").get_fields()


def test_get_fields_no_extension_provided():
    with pytest.raises(NotImplementedError):
        DatasetParser.for_file("my_dataset").get_fields()


def test_transform_datatype_unknown_type():
    expected = None
    input_data = "definitely not a known data type"
    actual = DatasetParser.transform_data_type(input_data)
    assert actual == expected


@pytest.mark.parametrize(
    ("expected", "concrete_type"),
    [
        *[(Datatype.INTEGER, i) for i in KNOWN_INTEGER_TYPES],
        *[(Datatype.FLOAT, i) for i in KNOWN_FLOAT_TYPES],
        *[(Datatype.STRING, i) for i in KNOWN_STRING_TYPES],
        *[(Datatype.DATETIME, i) for i in KNOWN_DATETIME_TYPES],
        *[(Datatype.BOOLEAN, i) for i in KNOWN_BOOLEAN_TYPES],
    ],
)
def test_transform_datatype(expected: Datatype, concrete_type: str):
    actual = DatasetParser.transform_data_type(concrete_type)
    assert actual == expected



"""Tests for the DataDocMetadata class."""
from __future__ import annotations

import json
from copy import copy
from pathlib import Path, PurePath
from typing import TYPE_CHECKING
from uuid import UUID

import pytest
from datadoc_model import Enums
from datadoc_model.Enums import DatasetState
from datadoc_model.Model import DataDocDataSet, DataDocVariable, MetadataDocument

from datadoc.backend.datadoc_metadata import PLACEHOLDER_USERNAME, DataDocMetadata

from .utils import (
    TEST_EXISTING_METADATA_DIRECTORY,
    TEST_EXISTING_METADATA_FILE_NAME,
    TEST_PARQUET_FILEPATH,
    TEST_RESOURCES_DIRECTORY,
)

if TYPE_CHECKING:
    from datetime import datetime


def make_paths() -> list[tuple[str, DatasetState]]:
    split_path = list(PurePath(TEST_PARQUET_FILEPATH).parts)
    initial_data = [
        ("kildedata", DatasetState.SOURCE_DATA),
        ("inndata", DatasetState.INPUT_DATA),
        ("roskildedata/klargjorte-data", DatasetState.PROCESSED_DATA),
        ("klargjorte_data", DatasetState.PROCESSED_DATA),
        ("klargjorte-data", DatasetState.PROCESSED_DATA),
        ("statistikk", DatasetState.STATISTIC),
        ("", None),
    ]
    test_data = []

    # Construct paths with each of the potential options in them
    for to_insert, state in initial_data:
        new_path = copy(split_path)
        new_path.insert(-2, to_insert)
        new_path = PurePath().joinpath(*new_path)
        test_data.append((str(new_path), state))

    return test_data


@pytest.mark.parametrize(("path", "expected_result"), make_paths())
def test_get_dataset_state(
    path: str,
    expected_result: DatasetState,
    metadata: DataDocMetadata,
):
    actual_state = metadata.get_dataset_state(path)
    assert actual_state == expected_result


def test_get_dataset_state_none(metadata: DataDocMetadata):
    assert metadata.get_dataset_state(None) is None


@pytest.mark.usefixtures("existing_metadata_file", "remove_document_file")
def test_existing_metadata_file(
    metadata: DataDocMetadata,
):
    assert metadata.meta.dataset.name.en == "successfully_read_existing_file"


def test_metadata_document_percent_complete(metadata: DataDocMetadata):
    dataset = DataDocDataSet(dataset_state=Enums.DatasetState.OUTPUT_DATA)
    variable_1 = DataDocVariable(data_type=Enums.Datatype.BOOLEAN)
    variable_2 = DataDocVariable(data_type=Enums.Datatype.INTEGER)
    document = MetadataDocument(
        percentage_complete=0,
        document_version=1,
        dataset=dataset,
        variables=[variable_1, variable_2],
    )
    metadata.meta = document

    assert metadata.percent_complete == 17  # noqa: PLR2004


def test_get_dataset_version(metadata: DataDocMetadata):
    assert metadata.get_dataset_version(metadata.short_name) == "1"


def test_get_dataset_version_unknown(metadata: DataDocMetadata):
    assert metadata.get_dataset_version("person_data.parquet") is None


@pytest.mark.usefixtures("remove_document_file")
def test_write_metadata_document(
    dummy_timestamp: datetime,
    metadata: DataDocMetadata,
):
    metadata.write_metadata_document()
    written_document = TEST_RESOURCES_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
    assert Path.exists(written_document)
    assert metadata.meta.dataset.metadata_created_date == dummy_timestamp
    assert metadata.meta.dataset.metadata_created_by == PLACEHOLDER_USERNAME
    assert metadata.meta.dataset.metadata_last_updated_date == dummy_timestamp
    assert metadata.meta.dataset.metadata_last_updated_by == PLACEHOLDER_USERNAME

    with Path.open(written_document) as f:
        written_metadata = json.loads(f.read())

    assert (
        # Use our pydantic model to read in the datetime string so we get the correct format
        DataDocDataSet(
            metadata_created_date=written_metadata["dataset"]["metadata_created_date"],
        ).metadata_created_date
        == dummy_timestamp
    )
    assert written_metadata["dataset"]["metadata_created_by"] == PLACEHOLDER_USERNAME
    assert (
        # Use our pydantic model to read in the datetime string so we get the correct format
        DataDocDataSet(
            metadata_last_updated_date=written_metadata["dataset"][
                "metadata_last_updated_date"
            ],
        ).metadata_last_updated_date
        == dummy_timestamp
    )
    assert (
        written_metadata["dataset"]["metadata_last_updated_by"] == PLACEHOLDER_USERNAME
    )


@pytest.mark.usefixtures("existing_metadata_file", "remove_document_file")
def test_write_metadata_document_existing_document(
    dummy_timestamp: datetime,
    metadata: DataDocMetadata,
):
    original_created_date: datetime = metadata.meta.dataset.metadata_created_date
    original_created_by = metadata.meta.dataset.metadata_created_by
    metadata.write_metadata_document()
    assert metadata.meta.dataset.metadata_created_by == original_created_by
    assert metadata.meta.dataset.metadata_created_date == original_created_date
    assert metadata.meta.dataset.metadata_last_updated_by == PLACEHOLDER_USERNAME
    assert metadata.meta.dataset.metadata_last_updated_date == dummy_timestamp


def test_metadata_id(metadata: DataDocMetadata):
    assert isinstance(metadata.meta.dataset.id, UUID)


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "invalid_id_field"],
)
@pytest.mark.usefixtures("remove_document_file")
def test_existing_metadata_none_id(
    existing_metadata_file: str,
    metadata: DataDocMetadata,
):
    pre_open_id = ""
    post_write_id = ""
    with Path.open(existing_metadata_file) as f:
        pre_open_id = json.load(f)["dataset"]["id"]
    assert pre_open_id is None
    assert isinstance(metadata.meta.dataset.id, UUID)
    metadata.write_metadata_document()
    with Path.open(existing_metadata_file) as f:
        post_write_id = json.load(f)["dataset"]["id"]
    assert post_write_id == str(metadata.meta.dataset.id)


@pytest.mark.parametrize(
    "existing_metadata_path",
    [TEST_EXISTING_METADATA_DIRECTORY / "valid_id_field"],
)
@pytest.mark.usefixtures("remove_document_file")
def test_existing_metadata_valid_id(
    existing_metadata_file: str,
    metadata: DataDocMetadata,
):
    pre_open_id = ""
    post_write_id = ""
    with Path.open(existing_metadata_file) as f:
        pre_open_id = json.load(f)["dataset"]["id"]
    assert pre_open_id is not None
    assert isinstance(metadata.meta.dataset.id, UUID)
    assert str(metadata.meta.dataset.id) == pre_open_id
    metadata.write_metadata_document()
    with Path.open(existing_metadata_file) as f:
        post_write_id = json.load(f)["dataset"]["id"]
    assert post_write_id == pre_open_id


def test_variable_role_default_value(metadata: DataDocMetadata):
    assert all(
        v.variable_role == Enums.VariableRole.MEASURE.value
        for v in metadata.meta.variables
    )


def test_direct_person_identifying_default_value(metadata: DataDocMetadata):
    assert all(not v.direct_person_identifying for v in metadata.meta.variables)



"""Tests for the callbacks package."""
from __future__ import annotations

import random
from copy import deepcopy

import pytest
from datadoc_model.Enums import DatasetState, Datatype, SupportedLanguages
from datadoc_model.Model import DataDocDataSet, DataDocVariable, LanguageStrings

from datadoc import state
from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.frontend.callbacks.dataset import (
    accept_dataset_metadata_input,
    change_language_dataset_metadata,
    update_dataset_metadata_language,
    update_global_language_state,
)
from datadoc.frontend.callbacks.utils import (
    MetadataInputTypes,
    find_existing_language_string,
)
from datadoc.frontend.callbacks.variables import (
    accept_variable_metadata_input,
    update_variable_table_dropdown_options_for_language,
    update_variable_table_language,
)
from datadoc.frontend.fields.display_dataset import DISPLAYED_DROPDOWN_DATASET_ENUMS
from datadoc.frontend.fields.display_variables import VariableIdentifiers
from datadoc.tests.utils import TEST_PARQUET_FILEPATH

DATA_ORIGINAL = [
    {
        "short_name": "pers_id",
        "variable_role": None,
    },
]
DATA_VALID = [
    {
        "short_name": "pers_id",
        "variable_role": "IDENTIFIER",
    },
]
DATA_NONETYPE = [
    {"short_name": "pers_id", "variable_role": "IDENTIFIER", "name": None},
]
DATA_INVALID = [
    {
        "short_name": "pers_id",
        "variable_role": 3.1415,
    },
]
DATA_CLEAR_URI = [
    {"short_name": "pers_id", "variable_role": None, "definition_uri": ""},
]

ENGLISH_NAME = "English Name"
BOKMÅL_NAME = "Bokmål Name"
NYNORSK_NAME = "Nynorsk Name"

LANGUAGE_OBJECT = LanguageStrings(en=ENGLISH_NAME, nb=BOKMÅL_NAME)


@pytest.fixture()
def active_cell() -> dict[str, MetadataInputTypes]:
    return {"row": 1, "column": 1, "column_id": "short_name", "row_id": None}


def test_accept_variable_metadata_input_no_change_in_data(
    metadata: DataDocMetadata,
    active_cell: dict[str, MetadataInputTypes],
):
    state.metadata = metadata
    output = accept_variable_metadata_input(DATA_ORIGINAL, active_cell, DATA_ORIGINAL)
    assert output[0] == DATA_ORIGINAL
    assert output[1] is False
    assert output[2] == ""


def test_accept_variable_metadata_input_new_data(
    active_cell: dict[str, MetadataInputTypes],
):
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    output = accept_variable_metadata_input(DATA_VALID, active_cell, DATA_ORIGINAL)

    assert state.metadata.variables_lookup["pers_id"].variable_role == "IDENTIFIER"
    assert output[0] == DATA_VALID
    assert output[1] is False
    assert output[2] == ""


def test_accept_variable_metadata_clear_string(
    active_cell: dict[str, MetadataInputTypes],
):
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    output = accept_variable_metadata_input(DATA_CLEAR_URI, active_cell, DATA_ORIGINAL)

    assert state.metadata.variables_lookup["pers_id"].definition_uri is None
    assert output[0] == DATA_CLEAR_URI
    assert output[1] is False
    assert output[2] == ""


def test_accept_variable_metadata_input_incorrect_data_type(
    active_cell: dict[str, MetadataInputTypes],
):
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    previous_metadata = deepcopy(state.metadata.meta.variables)
    output = accept_variable_metadata_input(DATA_INVALID, active_cell, DATA_ORIGINAL)

    assert output[0] == DATA_ORIGINAL
    assert output[1] is True
    assert "validation error for DataDocVariable" in output[2]
    assert state.metadata.meta.variables == previous_metadata


def test_accept_dataset_metadata_input_new_data():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    output = accept_dataset_metadata_input(DatasetState.INPUT_DATA, "dataset_state")
    assert output[0] is False
    assert output[1] == ""
    assert state.metadata.meta.dataset.dataset_state == "INPUT_DATA"


def test_accept_dataset_metadata_input_incorrect_data_type():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    output = accept_dataset_metadata_input(3.1415, "dataset_state")
    assert output[0] is True
    assert "validation error for DataDocDataSet" in output[1]


def test_update_dataset_metadata_language_strings():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    state.metadata.meta.dataset.name = LANGUAGE_OBJECT
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    output = update_dataset_metadata_language()
    assert ENGLISH_NAME not in output
    assert BOKMÅL_NAME in output
    state.current_metadata_language = SupportedLanguages.ENGLISH
    output = update_dataset_metadata_language()
    assert ENGLISH_NAME in output
    assert BOKMÅL_NAME not in output


def test_update_dataset_metadata_language_enums():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    state.metadata.meta.dataset.dataset_state = DatasetState.PROCESSED_DATA
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    output = update_dataset_metadata_language()
    assert DatasetState.PROCESSED_DATA.language_strings.en not in output
    assert DatasetState.PROCESSED_DATA.language_strings.nb not in output
    assert DatasetState.PROCESSED_DATA.name in output
    state.current_metadata_language = SupportedLanguages.ENGLISH
    output = update_dataset_metadata_language()
    assert DatasetState.PROCESSED_DATA.language_strings.en not in output
    assert DatasetState.PROCESSED_DATA.language_strings.nb not in output
    assert DatasetState.PROCESSED_DATA.name in output


def test_find_existing_language_string_no_existing_strings():
    dataset_metadata = DataDocDataSet()
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL
    language_strings = find_existing_language_string(
        dataset_metadata,
        BOKMÅL_NAME,
        "name",
    )
    assert language_strings == LanguageStrings(nb=BOKMÅL_NAME)


def test_find_existing_language_string_pre_existing_strings():
    dataset_metadata = DataDocDataSet()
    dataset_metadata.name = LANGUAGE_OBJECT
    state.current_metadata_language = SupportedLanguages.NORSK_NYNORSK
    language_strings = find_existing_language_string(
        dataset_metadata,
        NYNORSK_NAME,
        "name",
    )
    assert language_strings == LanguageStrings(
        nb=BOKMÅL_NAME,
        en=ENGLISH_NAME,
        nn=NYNORSK_NAME,
    )


def test_update_variable_table_language():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    test_variable = random.choice(  # noqa: S311 not for cryptographic purposes
        [v.short_name for v in state.metadata.meta.variables],
    )
    state.metadata.variables_lookup[test_variable].name = LANGUAGE_OBJECT
    output = update_variable_table_language(
        SupportedLanguages.NORSK_BOKMÅL,
    )
    try:
        name = next(
            d[VariableIdentifiers.NAME.value]
            for d in output[0]
            if d[VariableIdentifiers.SHORT_NAME.value] == test_variable
        )
    except StopIteration as e:
        msg = f"Could not find name for {test_variable = } in {output = }"
        raise AssertionError(msg) from e
    assert name == BOKMÅL_NAME


def test_nonetype_value_for_language_string(active_cell: dict[str, MetadataInputTypes]):
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    state.metadata.variables_lookup["pers_id"].name = LANGUAGE_OBJECT
    state.current_metadata_language = SupportedLanguages.NORSK_NYNORSK
    accept_variable_metadata_input(DATA_NONETYPE, active_cell, DATA_ORIGINAL)

    assert state.metadata.variables_lookup["pers_id"].name == LANGUAGE_OBJECT


def test_update_variable_table_dropdown_options_for_language():
    options = update_variable_table_dropdown_options_for_language(
        SupportedLanguages.NORSK_BOKMÅL,
    )
    assert all(k in DataDocVariable.__fields__ for k in options)
    assert all(list(v.keys()) == ["options"] for v in options.values())
    try:
        assert all(
            list(d.keys()) == ["label", "value"]
            for v in options.values()
            for d in next(iter(v.values()))
        )
    except StopIteration as e:
        msg = f"Could not extract actual value from {options.values() = }"
        raise AssertionError(msg) from e
    assert [d["label"] for d in options["data_type"]["options"]] == [
        i.get_value_for_language(SupportedLanguages.NORSK_BOKMÅL) for i in Datatype
    ]


def test_update_global_language_state():
    language: SupportedLanguages = (
        random.choice(  # noqa: S311 not for cryptographic purposes
            list(SupportedLanguages),
        )
    )
    update_global_language_state(language)
    assert state.current_metadata_language == language


def test_change_language_dataset_metadata():
    state.metadata = DataDocMetadata(str(TEST_PARQUET_FILEPATH))
    value = change_language_dataset_metadata(SupportedLanguages.NORSK_NYNORSK)
    test = random.choice(  # noqa: S311 not for cryptographic purposes
        DISPLAYED_DROPDOWN_DATASET_ENUMS,
    )
    assert isinstance(value, tuple)

    for options in value[0:-1]:
        assert all(list(d.keys()) == ["label", "value"] for d in options)

        member_names = set(test._member_names_)  # noqa: SLF001
        values = [i for d in options for i in d.values()]

        if member_names.intersection(values):
            assert {d["label"] for d in options} == {
                e.get_value_for_language(SupportedLanguages.NORSK_NYNORSK) for e in test
            }
            assert {d["value"] for d in options} == {e.name for e in test}



"""Shared fixtures and configuration."""

import shutil
import traceback
from datetime import datetime, timezone
from pathlib import Path
from unittest import mock

import pytest
from datadoc_model.Enums import SupportedLanguages
from pytest_mock import MockerFixture

from datadoc import state
from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.backend.storage_adapter import StorageAdapter

from .utils import (
    TEST_BUCKET_PARQUET_FILEPATH,
    TEST_EXISTING_METADATA_DIRECTORY,
    TEST_EXISTING_METADATA_FILE_NAME,
    TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY,
    TEST_PARQUET_FILEPATH,
    TEST_RESOURCES_DIRECTORY,
    TEST_RESOURCES_METADATA_DOCUMENT,
)


@pytest.fixture()
def dummy_timestamp() -> datetime:
    return datetime(2022, 1, 1, tzinfo=timezone.utc)


@pytest.fixture()
def _mock_timestamp(mocker: MockerFixture, dummy_timestamp: datetime) -> None:
    mocker.patch(
        "datadoc.backend.datadoc_metadata.get_timestamp_now",
        return_value=dummy_timestamp,
    )


@pytest.fixture()
def metadata(_mock_timestamp: None) -> DataDocMetadata:
    return DataDocMetadata(str(TEST_PARQUET_FILEPATH))


@pytest.fixture()
def remove_document_file() -> None:
    yield None  # Dummy value, No need to return anything in particular here
    full_path = TEST_RESOURCES_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
    try:
        full_path.unlink()
    except FileNotFoundError as e:
        print("File not deleted on teardown, exception caught:")  # noqa: T201
        traceback.print_exception(type(e), e)


@pytest.fixture()
def existing_metadata_path() -> Path:
    return TEST_EXISTING_METADATA_DIRECTORY


@pytest.fixture()
def existing_metadata_file(existing_metadata_path: Path) -> str:
    # Setup by copying the file into the relevant directory
    shutil.copy(
        existing_metadata_path / TEST_EXISTING_METADATA_FILE_NAME,
        TEST_RESOURCES_METADATA_DOCUMENT,
    )
    return str(TEST_RESOURCES_METADATA_DOCUMENT)


@pytest.fixture()
@pytest.mark.parametrize(
    "existing_metadata_file",
    [TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY],
)
def existing_metadata_with_valid_id_file(existing_metadata_file: Path) -> Path:
    return existing_metadata_file


@pytest.fixture()
def _clear_state() -> None:
    """Global fixture, referred to in pytest.ini."""
    state.metadata = None
    state.current_metadata_language = SupportedLanguages.NORSK_BOKMÅL


@pytest.fixture()
def mock_gcsfs_open(mocker: MockerFixture):
    return mocker.patch("gcsfs.GCSFileSystem.open")


@pytest.fixture()
def mock_gcsfs_exists(mocker: MockerFixture):
    mock = mocker.patch("gcsfs.GCSFileSystem.exists")
    mock.return_value = True
    return mock


@pytest.fixture()
def mock_pathlib_write_text(mocker: MockerFixture):
    return mocker.patch("pathlib.Path.write_text")


@pytest.fixture()
def local_parquet_file(mock_pathlib_write_text: mock.patch):  # noqa: ARG001
    return StorageAdapter.for_path(str(TEST_PARQUET_FILEPATH))


@pytest.fixture()
def bucket_object_parquet_file(
    mock_gcsfs_open: mock.patch,  # noqa: ARG001
    mock_gcsfs_exists: mock.patch,  # noqa: ARG001
):
    return StorageAdapter.for_path(TEST_BUCKET_PARQUET_FILEPATH)



"""Tests for the ModelBackwardsCompatibility class."""

import json
from pathlib import Path

import pytest

from datadoc.backend.datadoc_metadata import DataDocMetadata
from datadoc.backend.model_backwards_compatibility import (
    UnknownModelVersionError,
    upgrade_metadata,
)

from .utils import TEST_COMPATIBILITY_DIRECTORY

BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES = [
    d for d in TEST_COMPATIBILITY_DIRECTORY.iterdir() if d.is_dir()
]
BACKWARDS_COMPATIBLE_VERSION_NAMES = [
    d.stem for d in BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES
]


def test_existing_metadata_current_model_version():
    current_model_version = "1.0.0"
    fresh_metadata = {"document_version": current_model_version}
    upgraded_metadata = upgrade_metadata(fresh_metadata, current_model_version)
    assert upgraded_metadata == fresh_metadata


def test_existing_metadata_unknown_model_version():
    current_model_version = "1.0.0"
    fresh_metadata = {"document_version": "0.27.65"}
    with pytest.raises(UnknownModelVersionError):
        upgrade_metadata(fresh_metadata, current_model_version)


@pytest.mark.parametrize(
    "existing_metadata_path",
    BACKWARDS_COMPATIBLE_VERSION_DIRECTORIES,
    ids=BACKWARDS_COMPATIBLE_VERSION_NAMES,
)
@pytest.mark.usefixtures("remove_document_file")
def test_backwards_compatibility(
    existing_metadata_file: str,
    metadata: DataDocMetadata,
):
    # Parameterise with all known backwards compatible versions
    with Path.open(existing_metadata_file) as f:
        file_metadata = json.loads(f.read())

    in_file_values = [
        v for v in file_metadata["dataset"].values() if v not in ["", None]
    ]
    read_in_values = json.loads(metadata.meta.dataset.json(exclude_none=True)).values()

    missing_values = [v for v in in_file_values if v not in read_in_values]
    if missing_values:
        msg = f"Some values were not successfully read in! {missing_values = }"
        raise AssertionError(
            msg,
        )



"""Tests for the StorageAdapter class."""

import pathlib

import pytest

from datadoc.backend.storage_adapter import GCSObject, LocalFile, StorageAdapter
from datadoc.tests.utils import TEST_BUCKET_PARQUET_FILEPATH, TEST_PARQUET_FILEPATH


@pytest.mark.parametrize(
    ("file", "expected_class"),
    [("local_parquet_file", LocalFile), ("bucket_object_parquet_file", GCSObject)],
)
def test_factory(
    file: StorageAdapter,
    expected_class: StorageAdapter,
    request: pytest.FixtureRequest,
):
    # Ugly pytest magic to get the actual fixture out
    file: StorageAdapter = request.getfixturevalue(file)
    assert isinstance(file, expected_class)


@pytest.mark.parametrize("file", ["local_parquet_file", "bucket_object_parquet_file"])
def test_open(file: str, request: pytest.FixtureRequest):
    # Ugly pytest magic to get the actual fixture out
    file: StorageAdapter = request.getfixturevalue(file)
    with file.open() as file_handle:
        assert file_handle.readable()


@pytest.mark.parametrize(
    ("file", "expected_parent"),
    [
        (
            "local_parquet_file",
            str(TEST_PARQUET_FILEPATH.parent.absolute()),
        ),
        (
            "bucket_object_parquet_file",
            "/".join(TEST_BUCKET_PARQUET_FILEPATH.split("/")[:-1]),
        ),
    ],
)
def test_parent(file: str, expected_parent: str, request: pytest.FixtureRequest):
    # Ugly pytest magic to get the actual fixture out
    file: StorageAdapter = request.getfixturevalue(file)
    assert file.parent() == expected_parent


@pytest.mark.parametrize(
    ("known_file", "expected"),
    [
        (
            "local_parquet_file",
            f"{TEST_PARQUET_FILEPATH}/extra",
        ),
        (
            "bucket_object_parquet_file",
            f"{TEST_BUCKET_PARQUET_FILEPATH}/extra",
        ),
    ],
)
def test_joinpath(
    known_file: str,
    expected: str,
    request: pytest.FixtureRequest,
):
    # Ugly pytest magic to get the actual fixture out
    actual_file: StorageAdapter = request.getfixturevalue(known_file)
    actual_file.joinpath("extra")
    assert pathlib.Path(actual_file.location) == pathlib.Path(expected)


@pytest.mark.parametrize(
    "known_file",
    ["local_parquet_file", "bucket_object_parquet_file"],
)
def test_exists(known_file: str, request: pytest.FixtureRequest):
    # Ugly pytest magic to get the actual fixture out
    actual_file: StorageAdapter = request.getfixturevalue(known_file)
    assert actual_file.exists()


def test_write_text_local_file(
    local_parquet_file: StorageAdapter,
    request: pytest.FixtureRequest,
):
    local_parquet_file.write_text("12345")
    mock = request.getfixturevalue("mock_pathlib_write_text")
    mock.assert_called_once_with("12345", encoding="utf-8")


# Currently no test for writing to GCS
# Attempts to mock are failing with TypeError: cannot set 'write' attribute of immutable type '_io.BufferedWriter'



"""Utility values and functions for tests."""

from pathlib import Path

TEST_BUCKET_PARQUET_FILEPATH = "gs://ssb-staging-dapla-felles-data-delt/datadoc/klargjorte_data/person_data_v1.parquet"

TEST_RESOURCES_DIRECTORY = Path("datadoc/tests/resources/")

TEST_PARQUET_FILEPATH = TEST_RESOURCES_DIRECTORY / "person_data_v1.parquet"
TEST_SAS7BDAT_FILEPATH = TEST_RESOURCES_DIRECTORY / "sasdata.sas7bdat"
TEST_PARQUET_GZIP_FILEPATH = TEST_RESOURCES_DIRECTORY / "person_data_v1.parquet.gzip"

TEST_EXISTING_METADATA_DIRECTORY = TEST_RESOURCES_DIRECTORY / "existing_metadata_file"
TEST_EXISTING_METADATA_FILE_NAME = "person_data_v1__DOC.json"
TEST_EXISTING_METADATA_FILEPATH = (
    TEST_EXISTING_METADATA_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
)
TEST_RESOURCES_METADATA_DOCUMENT = (
    TEST_RESOURCES_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
)

TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY = (
    TEST_EXISTING_METADATA_DIRECTORY / "valid_id_field"
)

TEST_COMPATIBILITY_DIRECTORY = TEST_EXISTING_METADATA_DIRECTORY / "compatibility"
